"""
Functions and classes related to case conversions in schemata.
"""
import re
from typing import Pattern

from marshmallow import Schema, post_dump, pre_load


def snake_to_lower_camel_case(string: str) -> str:
    """
    Converts the string s from snake_case to lowerCamelCase
    :param string: foo_bar
    :return: fooBar
    """
    # copied from https://marshmallow.readthedocs.io/en/stable/examples.html?highlight=case#inflection-camel-casing-keys
    parts = iter(string.split("_"))
    return next(parts) + "".join(i.title() for i in parts)


_lower_camel_word_boundary_pattern: Pattern = re.compile("[^A-Z][A-Z]")


def lower_camel_to_snake_case(string: str) -> str:
    """
    Converts the string s from lowerCamelCase to snake_case
    :param string: fooBar
    :return: foo_bar
    """
    result = string
    for lower_upper in _lower_camel_word_boundary_pattern.findall(string):
        result = result.replace(lower_upper, lower_upper[0] + "_" + lower_upper[1])
    return result.lower()


_upper_camel_word_boundary_pattern: Pattern = re.compile("[A-Z][^A-Z]")


def upper_camel_to_snake_case(string: str) -> str:
    """
    Converts the string s from UpperCamel to snake_case
    :param string: FooBar
    :return: foo_bar
    """
    result = string
    for upper_lower in _upper_camel_word_boundary_pattern.findall(string):
        result = result.replace(upper_lower, upper_lower[0].lower() + "_" + upper_lower[1])
    return result


def to_snake_case(string: str) -> str:
    """
    Converts any string that is either lowerCamel or UpperCamel to snake_case
    :param string: fooBar or FooBar
    :return: foo_bar
    """
    result = lower_camel_to_snake_case(string)
    result = upper_camel_to_snake_case(result)
    return result


class CaseConverterSchema(Schema):
    """
    Schema that uses snake case for its internal representation and lowerCamelCase for external representation unless
    explicitly specified otherwise by using the data_key attribute.
    """

    def on_bind_field(self, field_name, field_obj):
        """
        if a data_key is not explicitly specified, then use the field name as lowerCamelCase
        :param field_name:
        :param field_obj:
        :return:
        """
        field_obj.data_key = field_obj.data_key or to_snake_case(field_name)

    @pre_load
    # pylint:disable=unused-argument
    def convert_keys_to_snake_case(self, data, many, **kwargs):
        """
        Converts all the keys of data to snake_case.
        :param data:
        :param many:
        :param kwargs:
        :return:
        """
        for external_key in list(data.keys()):
            internal_key = to_snake_case(external_key)
            if internal_key != external_key and internal_key in self.fields:
                data[internal_key] = data.pop(external_key)
        return data

    @post_dump(pass_many=True)
    def convert_keys_to_lower_camel(self, data, many, **kwargs):
        """
        Converts all the internal keys to lowerCamel keys
        :param data:
        :param many:
        :param kwargs:
        :return:
        """
        if many:
            return [self.convert_keys_to_lower_camel(d, many=False, **kwargs) for d in data]
        for internal_key in list(data.keys()):
            if internal_key in self.fields:
                external_key = snake_to_lower_camel_case(internal_key)
                if external_key != internal_key:
                    data[external_key] = data.pop(internal_key)
        return data



from typing import Dict, Optional

import attr
import pytest  # type:ignore[import]
from marshmallow import fields, post_load

from bo4e.schemata.caseconverterschema import CaseConverterSchema, snake_to_lower_camel_case, to_snake_case


@attr.s(auto_attribs=True)
class MyClass:
    some_string: str
    an_integer: int
    something: Optional[dict]


class MySchema(CaseConverterSchema):
    some_string = fields.Str(required=True, allow_none=False)
    an_integer = fields.Integer(required=True, allow_none=False)
    something = fields.Dict(required=False, allow_none=True)

    @post_load
    def to_my_class(self, data, **kwargs) -> MyClass:
        return MyClass(**data)


class MySchemaWithDataKeys(CaseConverterSchema):
    """
    The data_key attribute always allows to override the default behaviour.
    """

    some_string = fields.Str(required=True, allow_none=False, data_key="stringy key")
    an_integer = fields.Integer(required=True, allow_none=False, data_key="intiBinty")
    something = fields.Dict(required=False, allow_none=True, data_key="Some_Thing")

    class Meta:
        model = MyClass

    @post_load
    def to_my_class(self, data, **kwargs) -> MyClass:
        return MyClass(**data)


class TestCaseConversion:
    """
    Test class to test the snake/UpperCamel/lowerCamel conversions
    """

    @pytest.mark.parametrize(
        "original, expected",
        [
            pytest.param("foo", "foo"),
            pytest.param("foo_bar", "foo_bar"),
            pytest.param("fooBar", "foo_bar"),
            pytest.param("FooBar", "foo_bar"),
            pytest.param("Foo", "foo"),
            pytest.param("bo4e", "bo4e"),
            pytest.param("Bo4e", "bo4e"),
        ],
    )
    def test_to_snake(self, original: str, expected: str):
        """
        Tests the to_snake_case function
        :param original: input string
        :param expected: expected result
        :return:
        """
        actual = to_snake_case(original)
        assert actual == expected

    @pytest.mark.parametrize(
        "original, expected",
        [
            pytest.param("foo", "foo"),
            pytest.param("foo_bar", "fooBar"),
        ],
    )
    def test_to_lower_camel(self, original: str, expected: str):
        """
        Tests the to_lower_camel function
        :param original: input string
        :param expected: expected result
        :return:
        """
        actual = snake_to_lower_camel_case(original)
        assert actual == expected

    @pytest.mark.parametrize(
        "json_dict, expected",
        [
            pytest.param(
                {"some_string": "asd", "an_integer": 17, "something": None},
                MyClass(something=None, some_string="asd", an_integer=17),
            ),
            pytest.param(
                {"someString": "asd", "anInteger": 17, "something": None},
                MyClass(something=None, some_string="asd", an_integer=17),
            ),
            pytest.param(
                {"SomeString": "asd", "AnInteger": 17, "Something": None},
                MyClass(something=None, some_string="asd", an_integer=17),
            ),
        ],
    )
    def test_field_binding_on_load(self, json_dict: Dict[str, object], expected: MyClass):
        """
        Tests that the binding using the LowerCamelCaseSchema works
        :return:
        """
        schema = MySchema()
        actual: MyClass = schema.load(json_dict)
        assert actual == expected

    @pytest.mark.parametrize(
        "json_dict, expected",
        [
            pytest.param(
                {"stringy key": "asd", "intiBinty": 17, "Some_Thing": None},
                MyClass(something=None, some_string="asd", an_integer=17),
            )
        ],
    )
    def test_field_binding_on_load_with_data_keys(self, json_dict: Dict[str, object], expected: MyClass):
        """
        Tests that the binding using the LowerCamelCaseSchema works and the data_key parameter is taken into account.
        :return:
        """
        schema = MySchemaWithDataKeys()
        actual: MyClass = schema.load(json_dict)
        assert actual == expected

    @pytest.mark.parametrize(
        "instance, expected_dump",
        [
            pytest.param(
                MyClass(something=None, some_string="asd", an_integer=17),
                {"someString": "asd", "anInteger": 17, "something": None},
            ),
        ],
    )
    def test_key_conversion_on_dump(self, instance: MyClass, expected_dump: Dict[str, object]):
        """
        Tests that the dump using the LowerCamelCaseSchema works
        :return:
        """
        schema = MySchema()
        actual_dump: MyClass = schema.dump(instance)
        assert actual_dump == expected_dump

    @pytest.mark.parametrize(
        "instance, expected_dump",
        [
            pytest.param(
                MyClass(something=None, some_string="asd", an_integer=17),
                {"stringy key": "asd", "intiBinty": 17, "Some_Thing": None},
            ),
        ],
    )
    def test_key_conversion_on_dump_with_data_keys(self, instance: MyClass, expected_dump: Dict[str, object]):
        """
        Tests that the dump using the LowerCamelCaseSchema works and the data_key argument is taken into account.
        :return:
        """
        schema = MySchemaWithDataKeys()
        actual_dump: MyClass = schema.dump(instance)
        assert actual_dump == expected_dump



"""
Contains Geschaeftspartner class
and corresponding marshmallow schema for de-/serialization
"""
# pylint: disable=too-many-instance-attributes, too-few-public-methods
from typing import List, Optional, Type

import attr
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.com.adresse import Adresse, AdresseSchema
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart


@attr.s(auto_attribs=True, kw_only=True)
class Geschaeftspartner(Geschaeftsobjekt):
    """
    Object containing information about a Geschaeftspartner
    """

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.GESCHAEFTSPARTNER)
    name1: str
    gewerbekennzeichnung: bool
    geschaeftspartnerrolle: List[Geschaeftspartnerrolle] = attr.ib(validator=attr.validators.instance_of(List))

    # optional attributes
    anrede: Anrede = attr.ib(default=None)
    name2: Optional[str] = attr.ib(default=None)
    name3: Optional[str] = attr.ib(default=None)
    hrnummer: Optional[str] = attr.ib(default=None)
    amtsgericht: Optional[str] = attr.ib(default=None)
    kontaktweg: List[Kontaktart] = attr.ib(default=[])
    umsatzsteuer_id: Optional[str] = attr.ib(default=None)
    glaeubiger_id: Optional[str] = attr.ib(default=None)
    e_mail_adresse: Optional[str] = attr.ib(default=None)
    website: Optional[str] = attr.ib(default=None)
    partneradresse: Adresse = attr.ib(default=None)


class GeschaeftspartnerSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Geschaeftspartner.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name: Type[Geschaeftspartner] = Geschaeftspartner

    # required attributes
    name1 = fields.Str()
    gewerbekennzeichnung = fields.Bool()
    geschaeftspartnerrolle = fields.List(EnumField(Geschaeftspartnerrolle))

    # optional attributes
    anrede = EnumField(Anrede, missing=None)
    name2 = fields.Str(missing=None)
    name3 = fields.Str(missing=None)
    hrnummer = fields.Str(missing=None)
    amtsgericht = fields.Str(missing=None)
    kontaktweg = fields.List(EnumField(Kontaktart), missing=None)
    umsatzsteuer_id = fields.Str(missing=None)
    glaeubiger_id = fields.Str(missing=None)
    e_mail_adresse = fields.Str(missing=None)
    website = fields.Str(missing=None)
    partneradresse = fields.Nested(AdresseSchema, missing=None)



from decimal import Decimal

import pytest  # type:ignore[import]

from bo4e.com.menge import Menge, MengeSchema
from bo4e.enum.mengeneinheit import Mengeneinheit


class TestMenge:
    def test_menge(self):
        """
        Test de-/serialisation of Menge (only has required attributes).
        """
        menge = Menge(wert=Decimal(3.41), einheit=Mengeneinheit.MWH)

        schema = MengeSchema()
        json_string = schema.dumps(menge, ensure_ascii=False)

        assert "3.41" in json_string
        assert "MWH" in json_string

        menge_deserialized = schema.loads(json_string)

        assert isinstance(menge_deserialized.wert, Decimal)
        assert menge_deserialized.wert == Decimal(3.41)
        assert isinstance(menge_deserialized.einheit, Mengeneinheit)
        assert menge_deserialized.einheit == Mengeneinheit.MWH

    def test_wrong_datatype(self):
        with pytest.raises(TypeError) as excinfo:
            _ = Menge(wert="3.14", einheit=Mengeneinheit.MWH)

        assert "wert" in str(excinfo.value)

    def test_missing_required_attribute(self):
        with pytest.raises(TypeError) as excinfo:
            _ = Menge(wert=Decimal(3.14))

        assert "missing 1 required" in str(excinfo.value)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Marktrolle(StrEnum):
    """
    Diese Rollen kann ein Marktteilnehmer einnehmen.
    """

    NB = "NB"  # Grid Operator
    LF = "LF"  # Supplier
    MSB = "MSB"  # Messstellenbetreiber
    DL = "DL"  # Dienstleister
    BKV = "BKV"  # Bilanzkreisverantwortlicher,
    BKO = "BKO"  # Bilanzkoordinator / Marktgebietsverantwortlicher
    UENB = "UENB"  # Übertragungsnetzbetreiber
    KUNDE_SELBST_NN = "KUNDE_SELBST_NN"  # Kunden, die Netznutzungsentgelte selbst zahlen
    MGV = "MGV"  # Marktgebietsverantwortlicher
    EIV = "EIV"  # Einsatzverantwortlicher
    RB = "RB"  # Registerbetreiber
    KUNDE = "KUNDE"
    INTERESSENT = "INTERESSENT"
    BTR = "BTR"  # Betreiber einer technischen Ressource



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import DocumentedStrEnum


class Anrede(DocumentedStrEnum):
    """
    Übersicht möglicher Anreden, z.B. eines Geschäftspartners.
    """

    HERR = "HERR", "Herr"
    FRAU = "FRAU", "Frau"
    EHELEUTE = "EHELEUTE", "Eheleute"
    FIRMA = "FIRMA", "Firma"
    INDIVIDUELL = "INDIVIDUELL", 'Individuell (z.B. "Profx")'



# pylint:disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class BoTyp(StrEnum):
    """
    Auflistung sämtlicher existierender Geschäftsobjekte.
    """

    ANGEBOT = "ANGEBOT"
    ANSPRECHPARTNER = "ANSPRECHPARTNER"
    AUSSCHREIBUNG = "AUSSCHREIUNG"
    ENERGIEMENGE = "ENERGIEMENGE"
    # It is practical to use the BoTyp Enum as discriminator in the database.
    # Therefore we added one additional entry for GESCHAEFTSOBJEKT
    # This is not defined by the documentation!
    GESCHAEFTSOBJEKT = "GESCHAEFTSOBJEKT"
    GESCHAEFTSPARTNER = "GESCHAEFTSPARTNER"
    KOSTEN = "KOSTEN"
    MARKTLOKATION = "MARKTLOKATION"
    MESSLOKATION = "MESSLOKATION"
    MARKTTEILNEHMER = "MARKTTEILNEHMER"
    NETZNUTZUNGSRECHNUNG = "NETZNUTZUNGSRECHNUNG"
    PREISBLATT = "PREISBLATT"
    PREISBLATTDIENSTLEISTUNG = "PREISBLATTDIENSTLEISTUNG"
    PREISBLATTKONZESSIONSABGABE = "PREISBLATTKONZESSIONSABGABE"
    PREISBLATTMESSUNG = "PREISBLATTMESSUNG"
    PREISBLATTNETZNUTZUNG = "PREISBLATTNETZNUTZUNG"
    PREISBLATTUMLAGEN = "PREISBLATTUMLAGEN"
    RECHNUNG = "RECHNUNG"
    REGIONALTARIF = "REGIONALTARIF"
    TARIFINFO = "TARIFINFO"
    TARIFKOSTEN = "TARIFKOSTEN"
    TARIFPREISBLATT = "TARIFPREISBLATT"
    VERTRAG = "VERTRAG"
    ZAEHLER = "ZAEHLER"
    ZEITREIHE = "ZEITREIHE"



"""
Contains Vertrag class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import List, Optional

import attr
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.bo.geschaeftspartner import Geschaeftspartner, GeschaeftspartnerSchema
from bo4e.com.unterschrift import Unterschrift, UnterschriftSchema
from bo4e.com.vertragskonditionen import Vertragskonditionen, VertragskonditionenSchema
from bo4e.com.vertragsteil import Vertragsteil, VertragsteilSchema
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsart import Vertragsart
from bo4e.enum.vertragsstatus import Vertragsstatus


# pylint: disable=unused-argument
def at_least_one_vertragsteil(instance, attribute, value):
    """
    Ensures that the Vertrag has at least one entry in the Vertragsteile list.
    """
    if len(value) == 0:
        raise ValueError("The Vertrag must have at least 1 Vertragsteil")


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Vertrag(Geschaeftsobjekt):
    """
    Modell für die Abbildung von Vertragsbeziehungen.
    Das Objekt dient dazu, alle Arten von Verträgen, die in der Energiewirtschaft Verwendung finden, abzubilden.
    """

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.VERTRAG)
    vertragsnummer: str = attr.ib(validator=attr.validators.instance_of(str))
    vertragsart: Vertragsart = attr.ib(validator=attr.validators.instance_of(Vertragsart))
    vertragsstatus: Vertragsstatus = attr.ib(validator=attr.validators.instance_of(Vertragsstatus))
    sparte: Sparte = attr.ib(validator=attr.validators.instance_of(Sparte))
    vertragsbeginn: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    vertragsende: datetime = attr.ib(validator=attr.validators.instance_of(datetime))

    vertragspartner1: Geschaeftspartner = attr.ib(validator=attr.validators.instance_of(Geschaeftspartner))
    vertragspartner2: Geschaeftspartner = attr.ib(validator=attr.validators.instance_of(Geschaeftspartner))
    vertragsteile: List[Vertragsteil] = attr.ib(validator=at_least_one_vertragsteil)

    # optional attributes
    beschreibung: Optional[str] = attr.ib(default=None)
    vertragskonditionen: Optional[Vertragskonditionen] = attr.ib(default=None)
    unterzeichnervp1: Optional[List[Unterschrift]] = attr.ib(default=None)
    unterzeichnervp2: Optional[List[Unterschrift]] = attr.ib(default=None)


class VertragSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Vertrag.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Vertrag

    # required attributes
    vertragsnummer = fields.Str()
    vertragsart = EnumField(Vertragsart)
    vertragsstatus = EnumField(Vertragsstatus)
    sparte = EnumField(Sparte)
    vertragsbeginn = fields.DateTime()
    vertragsende = fields.DateTime()

    vertragspartner1 = fields.Nested(GeschaeftspartnerSchema)
    vertragspartner2 = fields.Nested(GeschaeftspartnerSchema)
    vertragsteile = fields.Nested(VertragsteilSchema, many=True)

    # optional attributes
    beschreibung = fields.Str(missing=None)
    vertragskonditionen = fields.Nested(VertragskonditionenSchema, missing=None)
    unterzeichnervp1 = fields.Nested(UnterschriftSchema, missing=None, many=True)
    unterzeichnervp2 = fields.Nested(UnterschriftSchema, missing=None, many=True)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Themengebiet(StrEnum):
    """
    Über dieses ENUM kann eine thematische Zuordnung, beispielsweise eines Ansprechpartners, vorgenommen werden.
    """

    ALLGEMEINER_INFORMATIONSAUSTAUSCH = "ALLGEMEINER_INFORMATIONSAUSTAUSCH"  # Allgemeiner Informationsaustausch
    AN_UND_ABMELDUNG = "AN_UND_ABMELDUNG"  # An- und Abmeldung
    ANSPRECHPARTNER_ALLGEMEIN = "ANSPRECHPARTNER_ALLGEMEIN"  # Ansprechpartner Allgemein
    ANSPRECHPARTNER_BDEW_DVGW = "ANSPRECHPARTNER_BDEW_DVGW"  # Ansprechpartner BDEW/DVGW
    ANSPRECHPARTNER_IT_TECHNIK = "ANSPRECHPARTNER_IT_TECHNIK"  # Ansprechpartner IT/Technik
    BILANZIERUNG = "BILANZIERUNG"  # Bilanzierung
    BILANZKREISKOORDINATOR = "BILANZKREISKOORDINATOR"  # Bilanzkreiskoordinator
    BILANZKREISVERANTWORTLICHER = "BILANZKREISVERANTWORTLICHER"  # Bilanzkreisverantwortlicher
    DATENFORMATE_ZERTIFIKATE_VERSCHLUESSELUNGEN = (
        "DATENFORMATE_ZERTIFIKATE_VERSCHLUESSELUNGEN"  # Datenformate, Zertifikate, Verschlüsselungen
    )
    DEBITORENMANAGEMENT = "DEBITORENMANAGEMENT"  # Debitorenmanagement
    DEMAND_SIDE_MANAGEMENT = "DEMAND_SIDE_MANAGEMENT"  # Demand-Side-Management
    EDI_VEREINBARUNG = "EDI_VEREINBARUNG"  # EDI-Vereinbarung
    EDIFACT = "EDIFACT"  # EDIFACT
    ENERGIEDATENMANAGEMENT = "ENERGIEDATENMANAGEMENT"  # Energiedatenmanagement
    FAHRPLANMANAGEMENT = "FAHRPLANMANAGEMENT"  # Fahrplanmanagement
    ALOCAT = "ALOCAT"  # Format:ALOCAT
    APERAK = "APERAK"  # Format:APERAK
    CONTRL = "CONTRL"  # Format:CONTRL
    INVOIC = "INVOIC"  # Format:INVOIC
    MSCONS = "MSCONS"  # Format:MSCONS
    ORDERS = "ORDERS"  # Format:ORDERS
    ORDERSP = "ORDERSP"  # Format:ORDERSP
    REMADV = "REMADV"  # Format:REMADV
    UTILMD = "UTILMD"  # Format:UTILMD
    GABI = "GABI"  # GaBi Gas
    GELI = "GELI"  # GeLi Gas
    GERAETERUECKGABE = "GERAETERUECKGABE"  # Geräterückgabe
    GERAETEWECHSEL = "GERAETEWECHSEL"  # Gerätewechsel
    GPKE = "GPKE"  # GPKE
    INBETRIEBNAHME = "INBETRIEBNAHME"  # Inbetriebnahme
    KAPAZITAETSMANAGEMENT = "KAPAZITAETSMANAGEMENT"  # Kapazitätsmanagement
    KLAERFAELLE = "KLAERFAELLE"  # Klärfälle
    LASTGAENGE_RLM = "LASTGAENGE_RLM"  # Lastgänge RLM
    LIEFERANTENRAHMENVERTRAG = "LIEFERANTENRAHMENVERTRAG"  # Lieferantenrahmenvertrag
    LIEFERANTENWECHSEL = "LIEFERANTENWECHSEL"  # Lieferantenwechsel
    MABIS = "MABIS"  # MaBiS
    MAHNWESEN = "MAHNWESEN"  # Mahnwesen
    MARKTGEBIETSVERANTWORTLICHER = "MARKTGEBIETSVERANTWORTLICHER"  # Marktgebietsverantwortlicher
    MARKTKOMMUNIKATION = "MARKTKOMMUNIKATION"  # Marktkommunikation
    MEHR_MINDERMENGEN = "MEHR_MINDERMENGEN"  # Mehr- Mindermengen
    MSB_MDL = "MSB_MDL"  # MSB - MDL
    NETZABRECHNUNG = "NETZABRECHNUNG"  # Netzabrechnung
    NETZENTGELTE = "NETZENTGELTE"  # Netzentgelte
    NETZMANAGEMENT = "NETZMANAGEMENT"  # Netzmanagement
    RECHT = "RECHT"  # Recht
    REGULIERUNGSMANAGEMENT = "REGULIERUNGSMANAGEMENT"  # Regulierungsmanagement
    REKLAMATIONEN = "REKLAMATIONEN"  # Reklamationen
    SPERREN_ENTSPERREN_INKASSO = "SPERREN_ENTSPERREN_INKASSO"  # Sperren/Entsperren/Inkasso
    STAMMDATEN = "STAMMDATEN"  # Stammdaten
    STOERUNGSFAELLE = "STOERUNGSFAELLE"  # Übermittlung_Störungsfälle
    TECHNISCHE_FRAGEN = "TECHNISCHE_FRAGEN"  # Technische Fragen
    UMSTELLUNG_INVOIC = "UMSTELLUNG_INVOIC"  # Umstellung INVOIC
    VERSCHLUESSELUNG_SIGNATUR = "VERSCHLUESSELUNG_SIGNATUR"  # Verschlüsselung/Signatur
    VERTRAGSMANAGEMENT = "VERTRAGSMANAGEMENT"  # Vertragsmanagement
    VERTRIEB = "VERTRIEB"  # Vertrieb
    WIM = "WIM"  # WiM
    ZAEHLERSTAENDE_SLP = "ZAEHLERSTAENDE_SLP"  # Zählerstände SLP
    ZAHLUNGSVERKEHR = "ZAHLUNGSVERKEHR"  # Zahlungsverkehr
    ZUORDNUNGSVEREINBARUNG = "ZUORDNUNGSVEREINBARUNG"  # Zuordnungsvereinbarung



"""
Contains base class for all business objects
and corresponding marshmallow schema for de-/serialization
"""
# pylint: disable=unused-argument, too-few-public-methods
from typing import List, Optional, Type

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.externereferenz import ExterneReferenz, ExterneReferenzSchema
from bo4e.enum.botyp import BoTyp
from bo4e.schemata.caseconverterschema import CaseConverterSchema


def _create_empty_referenzen_list() -> List[ExterneReferenz]:
    """
    A method with a type hint to please mypy
    https://stackoverflow.com/a/61281305/10009545
    :return:
    """
    return []


@attr.s(auto_attribs=True, kw_only=True)
class Geschaeftsobjekt:
    """
    Base class for all business objects
    """

    # required attributes
    versionstruktur: str = attr.ib(default="2")
    bo_typ: BoTyp = attr.ib(default=BoTyp.GESCHAEFTSOBJEKT)

    # optional attributes
    externe_referenzen: Optional[List[ExterneReferenz]] = attr.ib(
        default=_create_empty_referenzen_list(), validator=attr.validators.instance_of(List)  # type:ignore[arg-type]
    )


class GeschaeftsobjektSchema(CaseConverterSchema):
    """
    This is an "abstract" class.
    All business objects schemata do inherit from this class.
    """

    # class_name is needed to use the correct schema for deserialization.
    class_name: Type[Geschaeftsobjekt] = Geschaeftsobjekt

    # required attributes
    versionstruktur = fields.String()
    bo_typ = EnumField(BoTyp)

    # optional attributes
    externe_referenzen = fields.List(fields.Nested(ExterneReferenzSchema), missing=None)

    @post_load
    def deserialize(self, data, **kwargs):
        """Deserialize JSON to python object."""
        return type(self).class_name(**data)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Zeiteinheit(StrEnum):
    """
    Auflistung möglicher Einheiten zur Verwendung in zeitbezogenen Angaben.
    """

    SEKUNDE = "SEKUNDE"  # Sekunde
    MINUTE = "MINUTE"  # Minute
    STUNDE = "STUNDE"  # Stunde
    VIERTEL_STUNDE = "VIERTEL_STUNDE"  # Viertelstunde
    TAG = "TAG"  # Tag
    WOCHE = "WOCHE"  # Woche
    MONAT = "MONAT"  # Monat
    QUARTAL = "QUARTAL"  # Quartal
    HALBJAHR = "HALBJAHR"  # Halbjahr
    JAHR = "JAHR"  # Jahr



"""
strenum contains an enum that inherits from the plain enum and string.
"""

from enum import Enum


# pylint: disable=too-few-public-methods
class StrEnum(str, Enum):
    """
    An enum that has string values.
    """

    # see https://docs.python.org/3/library/enum.html?highlight=strenum#others


# pylint: disable=too-few-public-methods
class DocumentedStrEnum(StrEnum):
    """
    A StrEnum that has docstrings attached to its members.
    Use it like this:
    class MyDocumentedEnum(DocumentedStrEnum):
    Foo = "Serialized Foo", "my good docstring of Foo"
    Bar = "Serialized Bar", "bar is not foo (this is a docstring)"
    Then wherever MyDocumentedEnum is used in BOs or COMs, the attribute will serialize as either "Serialized Foo"
    or "Serialized Bar". And inspect.getdocs(MyDocumentedEnum.Foo) will return the respective docstring.
    This is unittested.
    """

    # see https://stackoverflow.com/a/50473952/10009545
    def __new__(cls, value, doc=None):
        self = str.__new__(cls, value)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Mengeneinheit(StrEnum):
    """
    Einheit: Messgrößen, die per Messung oder Vorgabe ermittelt werden können.
    """

    W = "W"  # Watt
    WH = "WH"  # Wattstunde
    KW = "KW"  # Kilowatt
    KWH = "KWH"  # Kilowattstunde
    KVARH = "KVARH"  # Kilovarstunde
    MW = "MW"  # Megawatt
    MWH = "MWH"  # Megawattstunde
    STUECK = "STUECK"  # Stückzahl
    KUBIKMETER = "KUBIKMETER"  # Kubikmeter (Gas)
    STUNDE = "STUNDE"  # Stunde
    TAG = "TAG"  # Tage
    MONAT = "MONAT"  # Monat
    JAHR = "JAHR"  # Jahr
    PROZENT = "PROZENT"  # Prozent



import inspect
import re
from pathlib import Path
from typing import List, Optional, TypeVar

import pytest  # type:ignore[import]

from bo4e.enum import anrede
from bo4e.enum.anrede import Anrede
from bo4e.enum.strenum import StrEnum


class TestEnums:
    """
    A test class that checks the enum construction.
    """

    starts_with_whitespace_pattern = re.compile(r"^[\s\n]+")
    ends_with_whitespace_pattern = re.compile(r"[\s\n]+$")

    TEnum = TypeVar("TEnum", bound=StrEnum)

    @staticmethod
    def _get_all_enum_classes() -> List[TEnum]:
        """
        returns a list of all bo4e.enum classes
        """
        arbitrary_enum_module_path = Path(anrede.__file__)
        result = []
        for python_file in arbitrary_enum_module_path.parent.glob("*.py"):
            # don't ask me why. but it works.
            enum_module = __import__("bo4e.enum." + python_file.name.split(".")[0])
            for _, member in inspect.getmembers(enum_module.enum):
                if inspect.ismodule(member):
                    candidate = inspect.getmembers(member)[0][1]
                    if inspect.isclass(candidate):
                        result.append(candidate)
        return result

    @staticmethod
    def _get_class_doc(enum_class: TEnum) -> Optional[str]:
        """
        asserts that enum class is a class and returns the class' docstring
        """
        assert inspect.isclass(enum_class)
        return inspect.getdoc(enum_class)

    def test_enum_classes_docstrings(self):
        """
        Tests that the docstrings of the enum classes do not start with whitespace or blank lines.
        """
        all_enums = TestEnums._get_all_enum_classes()
        assert len(all_enums) > 100  # just to be sure we're not using the wrong directory or path
        for enum_class in all_enums:
            docstring = TestEnums._get_class_doc(enum_class)
            assert docstring is not None
            assert not TestEnums.starts_with_whitespace_pattern.match(docstring)
            assert not TestEnums.ends_with_whitespace_pattern.match(docstring)

    @pytest.mark.parametrize(
        "enum_member, expected_docstring",
        [
            pytest.param(Anrede.HERR, "Herr"),
            pytest.param(Anrede.INDIVIDUELL, 'Individuell (z.B. "Profx")'),
        ],
    )
    def test_enum_member_docstrings_explicitly(self, enum_member: TEnum, expected_docstring: Optional[str]):
        """
        Test the docstrings of the enum members explicitly.
        if the general approach (using DocumentedStrEnum) works for single members, it will also work for all enums,
        which are constructed similarly.
        """
        assert inspect.getdoc(enum_member) == expected_docstring

    def test_enum_members_are_all_documented(self):
        """
        The class docstrings are enforced using pylint but the docstrings of enum members are not covered by pylint.
        """
        all_enums = self._get_all_enum_classes()
        for enum_class in all_enums:
            class_docstring = TestEnums._get_class_doc(enum_class)
            for enum_member in enum_class:
                member_docstring = inspect.getdoc(enum_member)
                assert (
                    member_docstring != class_docstring
                ), f"Docstring of Enum member {enum_member} must not be the same as the class docstring"
                assert member_docstring is not None
                assert member_docstring != ""
                assert not TestEnums.starts_with_whitespace_pattern.match(member_docstring)
                assert not TestEnums.ends_with_whitespace_pattern.match(member_docstring)
            break  # proof this works just for the wip anrede



"""
Contains Rufnummer class
and corresponding marshmallow schema for de-/serialization
"""

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.rufnummernart import Rufnummernart


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Rufnummer(COM):
    """
    Contains information to call or fax someone
    """

    # required attributes
    nummerntyp: Rufnummernart = attr.ib(validator=attr.validators.in_(Rufnummernart))
    rufnummer: str = attr.ib(validator=attr.validators.instance_of(str))


class RufnummerSchema(COMSchema):
    """
    Schema for de-/serialization of Rufnummer.
    """

    # required attributes
    nummerntyp = EnumField(Rufnummernart)
    rufnummer = fields.Str()

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Rufnummer:
        """Deserialize JSON to Rufnummer object"""
        return Rufnummer(**data)



"""
Contains Ansprechpartner class
and corresponding marshmallow schema for de-/serialization
"""
import attr
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.bo.geschaeftspartner import Geschaeftspartner, GeschaeftspartnerSchema
from bo4e.com.adresse import Adresse, AdresseSchema
from bo4e.com.rufnummer import Rufnummer, RufnummerSchema
from bo4e.com.zustaendigkeit import Zustaendigkeit, ZustaendigkeitSchema
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.titel import Titel


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Ansprechpartner(Geschaeftsobjekt):
    """
    Object containing information about a Ansprechpartner
    """

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.ANSPRECHPARTNER)
    nachname: str
    geschaeftspartner: Geschaeftspartner

    # optional attributes
    anrede: Anrede = attr.ib(default=None)
    individuelle_anrede: str = attr.ib(default=None)
    titel: Titel = attr.ib(default=None)
    vorname: str = attr.ib(default=None)
    e_mail_adresse: str = attr.ib(default=None)
    kommentar: str = attr.ib(default=None)
    adresse: Adresse = attr.ib(default=None)
    rufnummer: Rufnummer = attr.ib(default=None)
    zustaendigkeit: Zustaendigkeit = attr.ib(default=None)


class AnsprechpartnerSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Ansprechpartner.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Ansprechpartner

    # required attributes
    nachname = fields.Str()
    geschaeftspartner = fields.Nested(GeschaeftspartnerSchema)

    # optional attributes
    anrede = EnumField(Anrede, missing=None)
    individuelle_anrede = fields.Str(missing=None)
    titel = EnumField(Titel, missing=None)
    vorname = fields.Str(missing=None)
    e_mail_adresse = fields.Str(missing=None)
    kommentar = fields.Str(missing=None)
    adresse = fields.Nested(AdresseSchema, missing=None)
    rufnummer = fields.Nested(RufnummerSchema, missing=None)
    zustaendigkeit = fields.Nested(ZustaendigkeitSchema, missing=None)



"""
Contains Katasteradresse class
and corresponding marshmallow schema for de-/serialization
"""
import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Katasteradresse(COM):
    """
    Dient der Adressierung über die Liegenschafts-Information.
    """

    gemarkung_flur: str
    flurstueck: str


class KatasteradresseSchema(COMSchema):
    """
    Schema for de-/serialization of Katasteradresse.
    """

    gemarkung_flur = fields.Str()
    flurstueck = fields.Str()

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Katasteradresse:
        """Deserialize JSON to Katasteradresse object"""
        return Katasteradresse(**data)



import json

import pytest  # type:ignore[import]

from bo4e.com.adresse import Adresse, AdresseSchema
from bo4e.enum.landescode import Landescode


class TestAddress:
    def test_serialization_strasse(self):
        """
        Test serialization with strasse und hausnummer
        and default value of landescode
        """
        with open(
            "./tests/test_data/test_data_adresse/test_data_adresse_only_required_fields.json",
            encoding="utf-8",
        ) as json_file:
            address_test_data = json.load(json_file)

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
            strasse=address_test_data["strasse"],
            hausnummer=address_test_data["hausnummer"],
        )

        schema = AdresseSchema()
        address_dict = schema.dump(a)

        assert address_dict["strasse"] == "Nördliche Münchner Straße"
        assert address_dict["hausnummer"] == "27A"
        assert address_dict["postleitzahl"] == "82031"
        assert address_dict["landescode"] == "DE"

        deserialized_address = schema.load(address_dict)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.strasse == "Nördliche Münchner Straße"
        assert deserialized_address.hausnummer == "27A"
        assert deserialized_address.postleitzahl == "82031"
        assert deserialized_address.landescode == Landescode.DE

    def test_serialization_only_postfach(self):
        """Test serialization with postfach"""
        address_test_data = {
            "postleitzahl": "82031",
            "ort": "Grünwald",
            "postfach": "10 64 38",
        }

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
            postfach=address_test_data["postfach"],
        )

        schema = AdresseSchema()
        address_json = schema.dumps(a, ensure_ascii=False)

        assert "10 64 38" in address_json
        assert "82031" in address_json
        assert "DE" in address_json

        deserialized_address = schema.loads(address_json)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.postfach == "10 64 38"
        assert deserialized_address.postleitzahl == "82031"
        assert deserialized_address.landescode == Landescode.DE

    def test_serialization_only_required_fields(self):
        """Test serialization with just postleitzahl und ort"""
        address_test_data = {
            "postleitzahl": "82031",
            "ort": "Grünwald",
        }

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
        )

        schema = AdresseSchema()
        address_json = schema.dumps(a, ensure_ascii=False)

        assert "Grünwald" in address_json
        assert "82031" in address_json

        deserialized_address = schema.loads(address_json)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.ort == "Grünwald"
        assert deserialized_address.postleitzahl == "82031"

    def test_serialization_only_required_fields_landescode_AT(self):
        with open(
            "./tests/test_data/test_data_adresse/test_data_adresse_only_required_fields.json",
            encoding="utf-8",
        ) as json_file:
            address_test_data = json.load(json_file)

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
            strasse=address_test_data["strasse"],
            hausnummer=address_test_data["hausnummer"],
            landescode=Landescode.AT,
        )

        schema = AdresseSchema()
        address_json = schema.dumps(a, ensure_ascii=False)
        deserialized_address = schema.loads(address_json)

        assert deserialized_address.landescode == Landescode.AT

    def test_deserialization(self):
        json_string = r"""{"strasse":"Getreidegasse",
                 "hausnummer":"9",
                 "ort":"Salzburg",
                 "postleitzahl":"5020",
                 "landescode":"AT"}"""

        schema = AdresseSchema()
        a: Adresse = schema.loads(json_string)
        assert a.landescode is Landescode.AT

    @pytest.mark.datafiles("./tests/test_data/test_data_adresse/test_data_adresse_missing_plz.json")
    def test_missing_required_attribute(self, datafiles):
        """
        Test for getting an error message if a required attribute is missing
        """
        with open(datafiles / "test_data_adresse_missing_plz.json", encoding="utf-8") as json_file:
            address_test_data = json.load(json_file)

        with pytest.raises(TypeError) as excinfo:
            _ = Adresse(
                ort=address_test_data["ort"],
                strasse=address_test_data["strasse"],
                hausnummer=address_test_data["hausnummer"],
            )

        assert "postleitzahl" in str(excinfo.value)

    @pytest.mark.parametrize(
        "address_test_data, expected",
        [
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "Nördliche Münchner Straße",
                    "hausnummer": None,
                    "postfach": None,
                },
                "Missing hausnummer",
                id="Missing hausnummer",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "",
                    "hausnummer": "27A",
                    "postfach": None,
                },
                "Missing strasse",
                id="Missing strasse",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "Nördliche Münchner Straße",
                    "hausnummer": "27A",
                    "postfach": "10 64 38",
                },
                "hausnummer OR postfach",
                id="Strasse, hausnummer and postfach",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": None,
                    "hausnummer": "27A",
                    "postfach": "10 64 38",
                },
                "hausnummer OR postfach",
                id="Postfach and hausnummer",
            ),
        ],
    )
    def test_strasse_xor_postfach(self, address_test_data, expected):
        with pytest.raises(ValueError) as excinfo:
            _ = Adresse(
                postleitzahl=address_test_data["postleitzahl"],
                ort=address_test_data["ort"],
                strasse=address_test_data["strasse"],
                hausnummer=address_test_data["hausnummer"],
                postfach=address_test_data["postfach"],
            )
        assert expected in str(excinfo.value)

    def test_serialization_of_non_german_address(self):
        """
        Minimal working example
        :return:
        """
        a = Adresse(
            postleitzahl="6413", ort="Wildermieming", strasse="Gerhardhof", hausnummer="1", landescode=Landescode.AT
        )
        assert a.landescode == Landescode.AT
        serialized_address = AdresseSchema().dumps(a)
        assert '"AT"' in serialized_address
        deserialized_address = AdresseSchema().loads(serialized_address)
        assert deserialized_address.landescode == Landescode.AT



"""
Contains Unterschrift class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime

import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Unterschrift(COM):
    """
    Modellierung einer Unterschrift, z.B. für Verträge, Angebote etc.
    """

    # required attributes
    name: str = attr.ib(validator=attr.validators.instance_of(str))

    # optional attributes
    ort: str = attr.ib(default=None)
    datum: datetime = attr.ib(default=None)


class UnterschriftSchema(COMSchema):
    """
    Schema for de-/serialization of Unterschrift.
    """

    # required attributes
    name = fields.String()

    # optional attributes
    ort = fields.String(missing=None)
    datum = fields.DateTime(missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Unterschrift:
        """Deserialize JSON to Unterschrift object"""
        return Unterschrift(**data)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Netzebene(StrEnum):
    """
    Auflistung möglicher Netzebenen innerhalb der Energiearten Strom und Gas.
    """

    NSP = "NSP"  # Niederspannung; Strom
    MSP = "MSP"  # Mittelspannung; Strom
    HSP = "HSP"  # Hochspannung; Strom
    HSS = "HSS"  # Hoechstspannung; Strom
    MSP_NSP_UMSP = "MSP_NSP_UMSP"  # MS/NS Umspannung; Strom
    HSP_MSP_UMSP = "HSP_MSP_UMSP"  # HS/MS Umspannung; Strom
    HSS_HSP_UMSP = "HSS_HSP_UMSP"  # HOES/HS Umspannung; Strom
    HD = "HD"  # Hochdruck; Gas
    MD = "MD"  # Mitteldruck; Gas
    ND = "ND"  # Niederdruck; Gas



"""
Contains Adresse class
and corresponding marshmallow schema for de-/serialization
"""

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.landescode import Landescode


# pylint: disable=unused-argument
def strasse_xor_postfach(instance, attribute, value):
    """
    An address is valid if it contains a postfach XOR (a strasse AND hausnummer).
    This functions checks for these conditions of a valid address.
    """
    if instance.strasse or instance.hausnummer:
        if instance.postfach:
            raise ValueError("Enter either strasse and hausnummer OR postfach.")
        if not instance.strasse:
            raise ValueError("Missing strasse to hausnummer.")
        if not instance.hausnummer:
            raise ValueError("Missing hausnummer to strasse.")


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Adresse(COM):
    """
    Contains an address that can be used for most purposes.
    """

    # required attributes
    postleitzahl: str = attr.ib(validator=attr.validators.instance_of(str))
    ort: str = attr.ib(validator=attr.validators.instance_of(str))

    # optional attributes
    strasse: str = attr.ib(default=None, validator=strasse_xor_postfach)
    hausnummer: str = attr.ib(default=None, validator=strasse_xor_postfach)
    postfach: str = attr.ib(default=None, validator=strasse_xor_postfach)
    adresszusatz: str = attr.ib(default=None)
    co_ergaenzung: str = attr.ib(default=None)
    landescode: Landescode = attr.ib(default=Landescode.DE)  # type:ignore


class AdresseSchema(COMSchema):
    """
    Schema for de-/serialization of Adresse.
    """

    # required attributes
    postleitzahl = fields.Str()
    ort = fields.Str()

    # optional attributes
    strasse = fields.Str(missing=None)
    hausnummer = fields.Str(missing=None)
    postfach = fields.Str(missing=None)
    adresszusatz = fields.Str(missing=None)
    co_ergaenzung = fields.Str(missing=None)
    landescode = EnumField(Landescode)

    # pylint: disable=no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> Adresse:
        """Deserialize JSON to Adresse object"""
        return Adresse(**data)



"""
Contains base class for all components
"""
import attr

from bo4e.schemata.caseconverterschema import CaseConverterSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class COM:
    """
    base class for all components
    """


class COMSchema(CaseConverterSchema):
    """
    This is a base class.
    All components objects schemata are inherited from this class.
    """



"""
Contains Dienstleistung class
and corresponding marshmallow schema for de-/serialization
"""

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Dienstleistung(COM):
    """
    Abbildung einer abrechenbaren Dienstleistung.
    """

    # required attributes
    dienstleistungstyp: Dienstleistungstyp = attr.ib(validator=attr.validators.in_(Dienstleistungstyp))
    bezeichnung: str = attr.ib(validator=attr.validators.instance_of(str))


class DienstleistungSchema(COMSchema):
    """
    Schema for de-/serialization of Dienstleistung.
    """

    # required attributes
    dienstleistungstyp = EnumField(Dienstleistungstyp)
    bezeichnung = fields.Str()

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Dienstleistung:
        """Deserialize JSON to Dienstleistung object"""
        return Dienstleistung(**data)



"""
Contains ExterenzeReferenz class
and corresponding marshmallow schema for de-/serialization
"""
import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class ExterneReferenz(COM):
    """
    Viele Datenobjekte weisen in unterschiedlichen Systemen eine eindeutige ID (Kundennummer, GP-Nummer etc.) auf.
    Beim Austausch von Datenobjekten zwischen verschiedenen Systemen ist es daher hilfreich,
    sich die eindeutigen IDs der anzubindenden Systeme zu merken.
    """

    # required attributes
    ex_ref_name: str
    ex_ref_wert: str


class ExterneReferenzSchema(COMSchema):
    """
    Schema for de-/serialization of ExterneReferenz.
    Inherits from Schema and JavaScriptMixin.
    """

    ex_ref_name = fields.Str()
    ex_ref_wert = fields.Str()

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> ExterneReferenz:
        """Deserialize JSON to ExterneReferenz object"""
        return ExterneReferenz(**data)



"""
Contains Marktlokation class
and corresponding marshmallow schema for de-/serialization
"""
import re

import attr
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.bo.geschaeftspartner import Geschaeftspartner, GeschaeftspartnerSchema
from bo4e.com.adresse import Adresse, AdresseSchema
from bo4e.com.geokoordinaten import Geokoordinaten, GeokoordinatenSchema
from bo4e.com.katasteradresse import Katasteradresse, KatasteradresseSchema
from bo4e.com.messlokationszuordnung import Messlokationszuordnung, MesslokationszuordnungSchema
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.gasqualitaet import Gasqualitaet
from bo4e.enum.gebiettyp import Gebiettyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte
from bo4e.enum.verbrauchsart import Verbrauchsart

_malo_id_pattern = re.compile(r"^[1-9][\d]{10}$")


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Marktlokation(Geschaeftsobjekt):
    """
    Object containing information about a Marktlokation
    """

    # pylint: disable=unused-argument, no-self-use
    def _validate_marktlokations_id(self, marktlokations_id_attribute, value):
        if not value:
            raise ValueError("The marktlokations_id must not be empty.")
        if not _malo_id_pattern.match(value):
            raise ValueError(f"The marktlokations_id '{value}' does not match {_malo_id_pattern.pattern}")
        expected_checksum = Marktlokation._get_checksum(value)
        actual_checksum = value[10:11]
        if expected_checksum != actual_checksum:
            # pylint: disable=line-too-long
            raise ValueError(
                f"The marktlokations_id '{value}' has checksum '{actual_checksum}' but '{expected_checksum}' was expected."
            )

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.MARKTLOKATION)
    marktlokations_id: str = attr.ib(validator=_validate_marktlokations_id)
    sparte: Sparte
    energierichtung: Energierichtung
    bilanzierungsmethode: Bilanzierungsmethode
    netzebene: Netzebene

    # optional attributes
    verbrauchsart: Verbrauchsart = attr.ib(default=None)
    unterbrechbar: bool = attr.ib(default=None)
    netzbetreibercodenr: str = attr.ib(default=None)
    gebietstyp: Gebiettyp = attr.ib(default=None)
    netzgebietsnr: str = attr.ib(default=None)
    bilanzierungsgebiet: str = attr.ib(default=None)
    grundversorgercodenr: str = attr.ib(default=None)
    gasqualitaet: Gasqualitaet = attr.ib(default=None)
    endkunde: Geschaeftspartner = attr.ib(default=None)
    zugehoerige_messlokation: Messlokationszuordnung = attr.ib(default=None)

    # only one of the following three optional attributes can be set
    lokationsadresse: Adresse = attr.ib(default=None)
    geoadresse: Geokoordinaten = attr.ib(default=None)
    katasterinformation: Katasteradresse = attr.ib(default=None)

    @lokationsadresse.validator
    @geoadresse.validator
    @katasterinformation.validator
    def validate_address_info(self, address_attribute, value):
        """Checks that there is one and only one valid adress given."""
        all_address_attributes = [
            self.lokationsadresse,
            self.geoadresse,
            self.katasterinformation,
        ]
        amount_of_given_address_infos = len([i for i in all_address_attributes if i is not None])
        if amount_of_given_address_infos != 1:
            raise ValueError("No or more than one address information is given.")

    @staticmethod
    def _get_checksum(malo_id: str) -> str:
        """
        Get the checksum of a marktlokations id.
        a) Quersumme aller Ziffern in ungerader Position
        b) Quersumme aller Ziffern auf gerader Position multipliziert mit 2
        c) Summe von a) und b) d) Differenz von c) zum nächsten Vielfachen von 10 (ergibt sich hier 10, wird die
           Prüfziffer 0 genommen
        https://bdew-codes.de/Content/Files/MaLo/2017-04-28-BDEW-Anwendungshilfe-MaLo-ID_Version1.0_FINAL.PDF
        :param self:
        :return: the checksum as string
        """
        odd_checksum: int = 0
        even_checksum: int = 0
        # start counting at 1 to be consistent with the above description
        # of "even" and "odd" but stop at tenth digit.
        for i in range(1, 11):
            digit = malo_id[i - 1 : i]
            if i % 2 - 1 == 0:
                odd_checksum += int(digit)
            else:
                even_checksum += 2 * int(digit)
        result: int = (10 - ((even_checksum + odd_checksum) % 10)) % 10
        return str(result)


class MarktlokationSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Marktlokation.
    Inherits from GeschaeftsobjektSchema.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Marktlokation

    # required attributes
    marktlokations_id = fields.Str()
    sparte = EnumField(Sparte)
    energierichtung = EnumField(Energierichtung)
    bilanzierungsmethode = EnumField(Bilanzierungsmethode)
    netzebene = EnumField(Netzebene)

    # optional attributes
    verbrauchsart = EnumField(Verbrauchsart, missing=None)
    unterbrechbar = fields.Bool(missing=None)
    netzbetreibercodenr = fields.Str(missing=None)
    gebietstyp = EnumField(Gebiettyp, missing=None)
    netzgebietsnr = fields.Str(missing=None)
    bilanzierungsgebiet = fields.Str(missing=None)
    grundversorgercodenr = fields.Str(missing=None)
    gasqualitaet = EnumField(Gasqualitaet, missing=None)
    endkunde = fields.Nested(GeschaeftspartnerSchema, missing=None)
    zugehoerige_messlokation = fields.List(fields.Nested(MesslokationszuordnungSchema), missing=None)

    # only one of the following three optional attributes can be set
    lokationsadresse = fields.Nested(AdresseSchema, missing=None)
    geoadresse = fields.Nested(GeokoordinatenSchema, missing=None)
    katasterinformation = fields.Nested(KatasteradresseSchema, missing=None)



"""
Contains Marktteilnehmer class
and corresponding marshmallow schema for de-/serialization
"""
import attr
from attr.validators import matches_re
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftspartner import Geschaeftspartner, GeschaeftspartnerSchema
from bo4e.enum.botyp import BoTyp
from bo4e.enum.marktrolle import Marktrolle
from bo4e.enum.rollencodetyp import Rollencodetyp


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Marktteilnehmer(Geschaeftspartner):
    """
    Objekt zur Aufnahme der Information zu einem Marktteilnehmer
    """

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.MARKTTEILNEHMER)
    marktrolle: Marktrolle
    rollencodenummer: str = attr.ib(validator=matches_re(r"^\d{13}$"))
    rollencodetyp: Rollencodetyp

    # optional attributes
    makoadresse: str = attr.ib(default=None)


class MarktteilnehmerSchema(GeschaeftspartnerSchema):
    """
    Schema for de-/serialization of Marktteilnehmer.
    """

    # class_name is needed to use the correct schema for deserialization.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Marktteilnehmer

    # required attributes
    marktrolle = EnumField(Marktrolle)
    rollencodenummer = fields.Str()
    rollencodetyp = EnumField(Rollencodetyp)

    # optional attributes
    makoadresse = fields.Str(missing=None)



"""
Contains Vertragsteil class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import Optional

import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema
from bo4e.com.menge import Menge, MengeSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Vertragsteil(COM):
    """
    Abbildung für einen Vertragsteil. Der Vertragsteil wird dazu verwendet,
    eine vertragliche Leistung in Bezug zu einer Lokation (Markt- oder Messlokation) festzulegen.
    """

    # required attributes
    vertragsteilbeginn: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    vertragsteilende: datetime = attr.ib(validator=attr.validators.instance_of(datetime))

    # optional attributes
    lokation: Optional[str] = attr.ib(default=None)
    vertraglich_fixierte_menge: Optional[Menge] = attr.ib(default=None)
    minimale_abnahmemenge: Optional[Menge] = attr.ib(default=None)
    maximale_abnahmemenge: Optional[Menge] = attr.ib(default=None)


class VertragsteilSchema(COMSchema):
    """
    Schema for de-/serialization of Vertragsteil.
    """

    # required attributes
    vertragsteilbeginn = fields.DateTime()
    vertragsteilende = fields.DateTime()

    # optional attributes
    lokation = fields.String(missing=None)
    vertraglich_fixierte_menge = fields.Nested(MengeSchema, missing=None)
    minimale_abnahmemenge = fields.Nested(MengeSchema, missing=None)
    maximale_abnahmemenge = fields.Nested(MengeSchema, missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Vertragsteil:
        """Deserialize JSON to Vertragsteil object"""
        return Vertragsteil(**data)



"""
Contains Vertragskonditionen class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema
from bo4e.com.zeitraum import Zeitraum, ZeitraumSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Vertragskonditionen(COM):
    """
    Abbildung für Vertragskonditionen. Die Komponente wird sowohl im Vertrag als auch im Tarif verwendet.
    """

    # optional attributes
    beschreibung: str = attr.ib(default=None)
    anzahl_abschlaege: Decimal = attr.ib(default=None)
    vertragslaufzeit: Zeitraum = attr.ib(default=None)
    kuendigungsfrist: Zeitraum = attr.ib(default=None)
    vertragsverlaengerung: Zeitraum = attr.ib(default=None)
    abschlagszyklus: Zeitraum = attr.ib(default=None)


class VertragskonditionenSchema(COMSchema):
    """
    Schema for de-/serialization of Vertragskonditionen.
    """

    # optional attributes
    beschreibung = fields.String(missing=None)
    anzahl_abschlaege = fields.Decimal(missing=None, as_string=True)
    vertragslaufzeit = fields.Nested(ZeitraumSchema, missing=None)
    kuendigungsfrist = fields.Nested(ZeitraumSchema, missing=None)
    vertragsverlaengerung = fields.Nested(ZeitraumSchema, missing=None)
    abschlagszyklus = fields.Nested(ZeitraumSchema, missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Vertragskonditionen:
        """Deserialize JSON to Vertragskonditionen object"""
        return Vertragskonditionen(**data)



"""
Contains Menge class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.mengeneinheit import Mengeneinheit


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Menge(COM):
    """
    Abbildung einer Menge mit Wert und Einheit.
    """

    # required attributes
    wert: Decimal = attr.ib(validator=attr.validators.instance_of(Decimal))
    einheit: Mengeneinheit = attr.ib(validator=attr.validators.instance_of(Mengeneinheit))


class MengeSchema(COMSchema):
    """
    Schema for de-/serialization of Menge.
    """

    # required attributes
    wert = fields.Decimal(as_string=True)
    einheit = EnumField(Mengeneinheit)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Menge:
        """Deserialize JSON to Menge object"""
        return Menge(**data)



"""
Contains Geokoordinaten class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

import attr
from marshmallow import fields, post_load

from bo4e.com.com import COM, COMSchema


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Geokoordinaten(COM):
    """
    This component provides the geo-coordinates for a location.
    """

    # = attr.ib() has to be there, to make the validator work
    breitengrad: Decimal = attr.ib(validator=attr.validators.instance_of(Decimal))
    laengengrad: Decimal = attr.ib(validator=attr.validators.instance_of(Decimal))


class GeokoordinatenSchema(COMSchema):
    """
    Schema for de-/serialization of Geokoordinaten.

    Standard json library can not serialize Decimal type.
    Therefore these information will be serialized as string.
    During the deserialisiation it will become a Decimal type again.
    Link to official documentain:
    https://marshmallow.readthedocs.io/en/latest/api_reference.html?highlight=function#marshmallow.fields.Decimal
    """

    breitengrad = fields.Decimal(as_string=True)
    laengengrad = fields.Decimal(as_string=True)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Geokoordinaten:
        """Deserialize JSON to Geokoordinaten object"""
        return Geokoordinaten(**data)



import json

from bo4e.bo.marktteilnehmer import Marktteilnehmer, MarktteilnehmerSchema
from bo4e.com.adresse import Adresse
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.marktrolle import Marktrolle
from bo4e.enum.rollencodetyp import Rollencodetyp


class TestMarktteilnehmer:
    def test_serialization(self):
        mt = Marktteilnehmer(
            # required attributes of Marktteilnehmer only
            marktrolle=Marktrolle.DL,
            rollencodenummer="9903916000000",
            rollencodetyp=Rollencodetyp.BDEW,
            # required attributes inherited from Geschaeftspartner
            name1="Netze BW GmbH",
            gewerbekennzeichnung=True,
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
            partneradresse=Adresse(
                strasse="Schelmenwasenstraße",
                hausnummer="15",
                postleitzahl="70567",
                ort="Stuttgart",
            ),
        )

        assert mt.versionstruktur == "2", "versionstruktur was not automatically set"
        assert mt.bo_typ == BoTyp.MARKTTEILNEHMER, "boTyp was not automatically set"

        schema = MarktteilnehmerSchema()

        json_string = schema.dumps(mt, ensure_ascii=False)
        json_dict = json.loads(json_string)

        # Test camelcase
        assert "boTyp" in json_dict
        assert "marktrolle" in json_dict

        deserialized_mt: Marktteilnehmer = schema.loads(json_string)

        assert mt.marktrolle is deserialized_mt.marktrolle
        # Test snakecase
        assert deserialized_mt.bo_typ is BoTyp.MARKTTEILNEHMER



"""
Contains Zustaendigkeit class
and corresponding marshmallow schema for de-/serialization
"""

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.themengebiet import Themengebiet


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Zustaendigkeit(COM):
    """
    Enthält die zeitliche Zuordnung eines Ansprechpartners zu Abteilungen und Zuständigkeiten.
    """

    # required attributes
    themengebiet: Themengebiet = attr.ib(validator=attr.validators.in_(Themengebiet))

    # optional attributes
    jobtitel: str = attr.ib(default=None)
    abteilung: str = attr.ib(default=None)


class ZustaendigkeitSchema(COMSchema):
    """
    Schema for de-/serialization of Zustaendigkeit.
    """

    # required attributes
    themengebiet = EnumField(Themengebiet)

    # optional attributes
    jobtitel = fields.Str(missing=None)
    abteilung = fields.Str(missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Zustaendigkeit:
        """Deserialize JSON to Zustaendigkeit object"""
        return Zustaendigkeit(**data)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


# pylint: disable=line-too-long
class Geraetetyp(StrEnum):
    """
    Auflistung möglicher abzurechnender Gerätetypen.
    """

    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"  # Wechselstromzähler
    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"  # Drehstromzähler
    ZWEIRICHTUNGSZAEHLER = "ZWEIRICHTUNGSZAEHLER"  # Zweirichtungszähler
    RLM_ZAEHLER = "RLM_ZAEHLER"  # RLM-Zähler
    BALGENGASZAEHLER = "BALGENGASZAEHLER"  # Balgengaszähler
    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"  # Maximumzähler (Schleppzähler)
    MULTIPLEXANLAGE = "MULTIPLEXANLAGE"  # Multiplexeranlage
    PAUSCHALANLAGE = "PAUSCHALANLAGE"  # Pauschalanlagen
    VERSTAERKERANLAGE = "VERSTAERKERANLAGE"  # Verstärkeranlage
    SUMMATIONSGERAET = "SUMMATIONSGERAET"  # Summationsgerät
    IMPULSGEBER = "IMPULSGEBER"  # Impulsgeber
    EDL_21_ZAEHLERAUFSATZ = "EDL_21_ZAEHLERAUFSATZ"  # EDL 21 Zähleraufsatz für Zähler
    VIER_QUADRANTEN_LASTGANGZAEHLER = "VIER_QUADRANTEN_LASTGANGZAEHLER"  # Vier-Quadranten-Lastgangzähler
    MENGENUMWERTER = "MENGENUMWERTER"  # Mengenumwerter
    STROMWANDLER = "STROMWANDLER"  # Stromwandler
    SPANNUNGSWANDLER = "SPANNUNGSWANDLER"  # Spannungs-Wandler
    KOMBIMESSWANDLER = "KOMBIMESSWANDLER"  # Kombimesswandler
    BLOCKSTROMWANDLER = "BLOCKSTROMWANDLER"  # Blockstromwandler
    DATENLOGGER = "DATENLOGGER"  # Datenlogger
    KOMMUNIKATIONSANSCHLUSS = "KOMMUNIKATIONSANSCHLUSS"  # Kommunikationsanschluss
    MODEM = "MODEM"  # Modem
    TELEKOMMUNIKATIONSEINRICHTUNG = "TELEKOMMUNIKATIONSEINRICHTUNG"  # vom Messstellenbetreiber beigestellte Telekommunikationseinrichtung (Telefonanschluss)
    DREHKOLBENGASZAEHLER = "DREHKOLBENGASZAEHLER"  # Drehkolbengaszähler
    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"  # Turbinenradgaszähler
    ULTRASCHALLZAEHLER = "ULTRASCHALLZAEHLER"  # Ultraschallzähler
    WIRBELGASZAEHLER = "WIRBELGASZAEHLER"  # Wirbelgaszähler
    MODERNE_MESSEINRICHTUNG = "MODERNE_MESSEINRICHTUNG"  # moderne Messeinrichtung
    INTELLIGENTES_MESSYSTEM = "INTELLIGENTES_MESSYSTEM"  # Intelligentes Messystem
    ELEKTRONISCHER_HAUSHALTSZAEHLER = "ELEKTRONISCHER_HAUSHALTSZAEHLER"  # elektronischer Haushaltszähler
    STEUEREINRICHTUNG = "STEUEREINRICHTUNG"  # Steuereinrichtung
    TARIFSCHALTGERAET = "TARIFSCHALTGERAET"  # Tarifschaltgerät
    RUNDSTEUEREMPFAENGER = "RUNDSTEUEREMPFAENGER"  # Rundsteuerempfänger
    OPTIONALE_ZUS_ZAEHLEINRICHTUNG = "OPTIONALE_ZUS_ZAEHLEINRICHTUNG"  # optionale zusätzliche Zähleinrichtung
    MESSWANDLERSATZ_IMS_MME = "MESSWANDLERSATZ_IMS_MME"  # Messwandlersatz Strom iMS und mME, NSP
    KOMBIMESSWANDLER_IMS_MME = "KOMBIMESSWANDLER_IMS_MME"  # Kombimesswandlersatz (Strom u. Spg) iMS und mME
    TARIFSCHALTGERAET_IMS_MME = "TARIFSCHALTGERAET_IMS_MME"  # Tarifschaltung iMS und mME
    RUNDSTEUEREMPFAENGER_IMS_MME = "RUNDSTEUEREMPFAENGER_IMS_MME"  # Rundsteuerempfänger iMS und mME
    TEMPERATUR_KOMPENSATION = "TEMPERATUR_KOMPENSATION"  # Temperaturkompensation
    HOECHSTBELASTUNGS_ANZEIGER = "HOECHSTBELASTUNGS_ANZEIGER"  # Höchsbelastungsanzeiger
    SONSTIGES_GERAET = "SONSTIGES_GERAET"  # Sonstiges Gerät
    PREPAYMENTZAEHLER = "PREPAYMENTZAEHLER"  # Prepaymentzähler
    EDL_21 = "EDL_21"  # EDL21
    _96_H_ZAEHLER = "96_H_ZAEHLER"  # 96 h Zähler
    EDL_40_ZAEHLERAUFSATZ = "EDL_40_ZAEHLERAUFSATZ"  # EDL 40 Zähleraufsatz für Zähler



"""
Contains Zeitraum class
and corresponding marshmallow schema for de-/serialization
"""

from datetime import datetime
from decimal import Decimal
from typing import Optional

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.zeiteinheit import Zeiteinheit


# pylint: disable=unused-argument
def time_range_possibilities(instance, attribute, value):
    """
    An address is valid if it contains a postfach XOR (a strasse AND hausnummer).
    This functions checks for these conditions of a valid address.
    """
    if (
        instance.einheit
        and instance.dauer
        and not (instance.startdatum or instance.enddatum or instance.startzeitpunkt or instance.endzeitpunkt)
    ):
        return
    if (
        instance.startdatum
        and instance.enddatum
        and not (instance.einheit or instance.dauer or instance.startzeitpunkt or instance.endzeitpunkt)
    ):
        return
    if (
        instance.startzeitpunkt
        and instance.endzeitpunkt
        and not (instance.einheit or instance.dauer or instance.startdatum or instance.enddatum)
    ):
        return

    raise ValueError(
        """
        Please choose from one of the three possibilities to specify the timerange:
        - einheit and dauer
        - startdatum and enddatum
        - startzeitpunkt and endzeitpunkt
        """
    )


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Zeitraum(COM):
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher eine der drei Möglichkeiten angegeben sein:
    - Einheit und Dauer oder
    - Zeitraum: Startdatum bis Enddatum oder
    - Zeitraum: Startzeitpunkt (Datum und Uhrzeit) bis Endzeitpunkt (Datum und Uhrzeit)
    """

    # optional attributes
    einheit: Optional[Zeiteinheit] = attr.ib(default=None, validator=time_range_possibilities)
    dauer: Optional[Decimal] = attr.ib(default=None, validator=time_range_possibilities)
    startdatum: Optional[datetime] = attr.ib(default=None, validator=time_range_possibilities)
    enddatum: Optional[datetime] = attr.ib(default=None, validator=time_range_possibilities)
    startzeitpunkt: Optional[datetime] = attr.ib(default=None, validator=time_range_possibilities)
    endzeitpunkt: Optional[datetime] = attr.ib(default=None, validator=time_range_possibilities)


class ZeitraumSchema(COMSchema):
    """
    Schema for de-/serialization of Zeitraum.
    """

    # optional attributes
    einheit = EnumField(Zeiteinheit, missing=None)
    dauer = fields.Decimal(missing=None, as_string=True)
    startdatum = fields.DateTime(missing=None)
    enddatum = fields.DateTime(missing=None)
    startzeitpunkt = fields.DateTime(missing=None)
    endzeitpunkt = fields.DateTime(missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Zeitraum:
        """Deserialize JSON to Zeitraum object"""
        return Zeitraum(**data)



"""
Contains Messlokationszuordnung class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime

import attr
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.arithmetische_operation import ArithmetischeOperation


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Messlokationszuordnung(COM):
    """
    Mit dieser Komponente werden Messlokationen zu Marktlokationen zugeordnet.
    Dabei kann eine arithmetische Operation (Addition, Subtraktion, Multiplikation, Division) angegeben werden,
    mit der die Messlokation zum Verbrauch der Marktlokation beiträgt.
    """

    # required attributes
    messlokations_id: str
    arithmetik: ArithmetischeOperation

    # optional attributes
    gueltig_seit: datetime = attr.ib(default=None)
    gueltig_bis: datetime = attr.ib(default=None)


class MesslokationszuordnungSchema(COMSchema):
    """
    Schema for de-/serialization of Katasteradresse.
    """

    # required attributes
    messlokations_id = fields.Str()
    arithmetik = EnumField(ArithmetischeOperation)

    # optional attributes
    gueltig_seit = fields.DateTime(missing=None)
    gueltig_bis = fields.DateTime(missing=None)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Messlokationszuordnung:
        """Deserialize JSON to Messlokationszuordnung object"""
        return Messlokationszuordnung(**data)



"""
Contains Zaehler class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from decimal import Decimal
from typing import List, Optional

import attr
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.bo.geschaeftspartner import Geschaeftspartner, GeschaeftspartnerSchema
from bo4e.com.zaehlwerk import Zaehlwerk, ZaehlwerkSchema
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.zaehlerauspraegung import Zaehlerauspraegung
from bo4e.enum.zaehlertyp import Zaehlertyp


# pylint: disable=unused-argument
def at_least_one_zaehlwerk(instance, attribute, value):
    """
    Ensures that the Zaehler has at least one entry in the zaehlwerke list.
    """
    if len(value) == 0:
        raise ValueError("The Zaehler must have at least 1 Zaehlwerk")


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Zaehler(Geschaeftsobjekt):
    """
    Object containing information about a meter/"Zaehler".
    """

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.ZAEHLER)
    zaehlernummer: str = attr.ib(
        validator=attr.validators.instance_of(str)
    )  # Nummerierung des Zählers, vergeben durch den Messstellenbetreiber
    sparte: Sparte
    zaehlerauspraegung: Zaehlerauspraegung
    zaehlertyp: Zaehlertyp
    zaehlwerke: List[Zaehlwerk] = attr.ib(validator=at_least_one_zaehlwerk)
    tarifart: Tarifart  # Spezifikation bezüglich unterstützter Tarifarten

    # optional attributes
    zaehlerkonstante: Optional[Decimal] = attr.ib(default=None)  # Zählerkonstante auf dem Zähler
    eichung_bis: Optional[datetime] = attr.ib(default=None)  # Bis zu diesem Datum ist der Zähler geeicht.
    letzte_eichung: Optional[datetime] = attr.ib(
        default=None
    )  # Zu diesem Datum fand die letzte Eichprüfung des Zählers statt.
    zaehlerhersteller: Optional[Geschaeftspartner] = attr.ib(default=None)  # Der Hersteller des Zählers.


class ZaehlerSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Zaehler.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Zaehler

    # required attributes
    zaehlernummer = fields.Str()
    sparte = EnumField(Sparte)
    zaehlerauspraegung = EnumField(Zaehlerauspraegung)
    zaehlertyp = EnumField(Zaehlertyp)
    tarifart = EnumField(Tarifart)
    zaehlwerke = fields.Nested(ZaehlwerkSchema, many=True)

    # optional attributes
    zaehlerkonstante = fields.Decimal(missing=None, as_string=True)
    eichung_bis = fields.DateTime(missing=None)
    letzte_eichung = fields.DateTime(missing=None)

    zaehlerhersteller = fields.Nested(GeschaeftspartnerSchema, missing=None)



"""
Contains Messlokation class
and corresponding marshmallow schema for de-/serialization
"""
import re
from typing import List, Optional

import attr
from iso3166 import countries  # type:ignore[import]
from marshmallow import fields
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt, GeschaeftsobjektSchema
from bo4e.bo.zaehler import Zaehler, ZaehlerSchema
from bo4e.com.adresse import Adresse, AdresseSchema
from bo4e.com.dienstleistung import Dienstleistung, DienstleistungSchema
from bo4e.com.geokoordinaten import Geokoordinaten, GeokoordinatenSchema
from bo4e.com.hardware import Hardware, HardwareSchema
from bo4e.com.katasteradresse import Katasteradresse, KatasteradresseSchema
from bo4e.enum.botyp import BoTyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte

# Structure of a Messlokations-ID
# Ländercode nach DIN ISO 3166 (2 Stellen)
# Verteilnetzbetreiber (6 Stellen)
# Postleitzahl (5 Stellen)
# Zählpunktnummer (20 Stellen alphanumerisch)
# source: https://de.wikipedia.org/wiki/Z%C3%A4hlpunkt#Struktur_der_Z%C3%A4hlpunktbezeichnung

_melo_id_pattern = re.compile(r"^[A-Z]{2}\d{6}\d{5}[A-Z\d]{20}$")


# pylint: disable=too-many-instance-attributes, too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Messlokation(Geschaeftsobjekt):
    """
    Object containing information about a Messlokation
    """

    # pylint: disable=unused-argument, no-self-use
    def _validate_messlokations_id(self, messlokations_id_attribute, value):
        if not value:
            raise ValueError("The messlokations_id must not be empty.")
        if not _melo_id_pattern.match(value):
            raise ValueError(f"The messlokations_id '{value}' does not match {_melo_id_pattern.pattern}")
        if not value[0:2] in countries:
            raise ValueError(f"The country code '{value[0:2]}' is not a valid country code")

    # required attributes
    bo_typ: BoTyp = attr.ib(default=BoTyp.MESSLOKATION)
    messlokations_id: str = attr.ib(validator=_validate_messlokations_id)
    sparte: Sparte

    # optional attributes
    netzebene_messung: Optional[Netzebene] = attr.ib(default=None)
    messgebietnr: Optional[str] = attr.ib(default=None)
    geraete: Optional[List[Hardware]] = attr.ib(default=None)
    messdienstleistung: Optional[List[Dienstleistung]] = attr.ib(default=None)
    messlokationszaehler: Optional[List[Zaehler]] = attr.ib(default=None)

    # only one of the following two optional codenr attributes can be set
    grundzustaendiger_msb_codenr: Optional[str] = attr.ib(default=None)
    grundzustaendiger_msbim_codenr: Optional[str] = attr.ib(default=None)

    # only one of the following three optional address attributes can be set
    messadresse: Optional[Adresse] = attr.ib(default=None)
    geoadresse: Optional[Geokoordinaten] = attr.ib(default=None)
    katasterinformation: Optional[Katasteradresse] = attr.ib(default=None)

    @messadresse.validator
    @geoadresse.validator
    @katasterinformation.validator
    def validate_address_info(self, address_attribute, value):
        """Checks that if an address is given, that there is only one valid address given"""
        all_address_attributes = [
            self.messadresse,
            self.geoadresse,
            self.katasterinformation,
        ]
        amount_of_given_address_infos = len([i for i in all_address_attributes if i is not None])
        if amount_of_given_address_infos > 1:
            raise ValueError("More than one address information is given.")

    @grundzustaendiger_msb_codenr.validator
    @grundzustaendiger_msbim_codenr.validator
    def validate_grundzustaendiger_x_codenr(self, attribute, value):
        """Checks that if a codenr is given, that there is only one valid codenr given."""
        all_grundzustaendiger_x_codenr_attributes = [
            self.grundzustaendiger_msb_codenr,
            self.grundzustaendiger_msbim_codenr,
        ]
        amount_of_given_grundzustaendiger_x_codenr = len(
            [i for i in all_grundzustaendiger_x_codenr_attributes if i is not None]
        )
        if amount_of_given_grundzustaendiger_x_codenr > 1:
            raise ValueError("More than one codenr is given.")


class MesslokationSchema(GeschaeftsobjektSchema):
    """
    Schema for de-/serialization of Marktlokation.
    Inherits from GeschaeftsobjektSchema.
    """

    # class_name is needed to use the correct schema for deserialisation.
    # see function `deserialize` in geschaeftsobjekt.py
    class_name = Messlokation

    # required attributes
    messlokations_id = fields.Str()
    sparte = EnumField(Sparte)

    # optional attributes
    netzebene_messung = EnumField(Netzebene, missing=None)
    messgebietnr = fields.Str(missing=None)
    geraete = fields.List(fields.Nested(HardwareSchema), missing=None)  #: List[Hardware]
    messdienstleistung = fields.List(fields.Nested(DienstleistungSchema), missing=None)  #: List[Dienstleistung]
    messlokationszaehler = fields.List(fields.Nested(ZaehlerSchema), missing=None)

    # only one of the following two optional codenr attributes can be set
    grundzustaendiger_msb_codenr = fields.Str(missing=None)
    grundzustaendiger_msbim_codenr = fields.Str(missing=None)

    # only one of the following three optional attributes can be set
    messadresse = fields.Nested(AdresseSchema, missing=None)
    geoadresse = fields.Nested(GeokoordinatenSchema, missing=None)
    katasterinformation = fields.Nested(KatasteradresseSchema, missing=None)



from datetime import datetime, timezone

import pytest  # type:ignore[import]

from bo4e.com.unterschrift import Unterschrift, UnterschriftSchema


class TestUnterschrift:
    def test_unterschrift_only_required_attributes(self):
        """
        Test de-/serialisation of Unterschrift with minimal attributes.
        """
        unterschrift = Unterschrift(name="Foo")

        schema = UnterschriftSchema()
        json_string = schema.dumps(unterschrift, ensure_ascii=False)

        assert "Foo" in json_string

        unterschrift_deserialized = schema.loads(json_string)

        assert isinstance(unterschrift_deserialized.name, str)
        assert unterschrift_deserialized.name == "Foo"

    def test_unterschrift_required_and_optional_attributes(self):
        """
        Test de-/serialisation of Unterschrift with maximal attributes.
        """
        unterschrift = Unterschrift(name="Foo", ort="Grünwald", datum=datetime(2019, 6, 7, tzinfo=timezone.utc))

        schema = UnterschriftSchema()
        json_string = schema.dumps(unterschrift, ensure_ascii=False)

        assert "Foo" in json_string
        assert "Grünwald" in json_string
        assert "2019-06-07T00:00:00+00:00" in json_string

        unterschrift_deserialized = schema.loads(json_string)

        assert isinstance(unterschrift_deserialized.name, str)
        assert unterschrift_deserialized.name == "Foo"
        assert isinstance(unterschrift_deserialized.ort, str)
        assert unterschrift_deserialized.ort == "Grünwald"
        assert isinstance(unterschrift_deserialized.datum, datetime)
        assert unterschrift_deserialized.datum == datetime(2019, 6, 7, tzinfo=timezone.utc)

    def test_unterschrift_missing_required_attribute(self):
        with pytest.raises(TypeError) as excinfo:
            _ = Unterschrift()

        assert "missing 1 required" in str(excinfo.value)



"""
Contains Zaehlwerk class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

import attr
from attr.validators import matches_re
from marshmallow import fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from bo4e.com.com import COM, COMSchema
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.mengeneinheit import Mengeneinheit


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class Zaehlwerk(COM):
    """
    Mit dieser Komponente werden Zählwerke modelliert.
    """

    zaehlwerk_id: str = attr.ib(
        validator=attr.validators.instance_of(str)
    )  # Identifikation des Zählwerks (Registers) innerhalb des Zählers.
    # Oftmals eine laufende Nummer hinter der Zählernummer. Z.B. 47110815_1
    bezeichnung: str = attr.ib(
        validator=attr.validators.instance_of(str)
    )  # Zusätzliche Bezeichnung, z.B. Zählwerk_Wirkarbeit.
    richtung: Energierichtung  # Die Energierichtung, Einspeisung oder Ausspeisung.
    obis_kennzahl: str = attr.ib(
        validator=matches_re(
            r"(?:(1)-((?:[0-5]?[0-9])|(?:6[0-5])):((?:[1-8]|99))\.((?:6|8|9|29))\.([0-9]{1,2}))|"
            r"(?:(7)-((?:[0-5]?[0-9])|(?:6[0-5])):(.{1,2})\.(.{1,2})\.([0-9]{1,2}))"
        )
    )  # Die OBIS-Kennzahl für das Zählwerk, die festlegt, welche auf die gemessene Größe mit dem Stand gemeldet wird.
    # Nur Zählwerkstände mit dieser OBIS-Kennzahl werden an diesem Zählwerk registriert.
    wandlerfaktor: Decimal = attr.ib(
        validator=attr.validators.instance_of(Decimal)
    )  # Mit diesem Faktor wird eine Zählerstandsdifferenz multipliziert, um zum eigentlichen Verbrauch im Zeitraum
    # zu kommen.
    einheit: Mengeneinheit  # Die Einheit der gemessenen Größe, z.B. kWh


class ZaehlwerkSchema(COMSchema):
    """
    Schema for de-/serialization of Zaehlwerk.
    """

    zaehlwerk_id = fields.Str()
    bezeichnung = fields.Str()
    richtung = EnumField(Energierichtung)
    obis_kennzahl = fields.Str()
    wandlerfaktor = fields.Decimal(as_string=True)
    einheit = EnumField(Mengeneinheit)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> Zaehlwerk:
        """Deserialize JSON to Zaehlwerk object"""
        return Zaehlwerk(**data)
